/*
    userver -- (pronounced you-server or micro-server).
    This file is part of the userver, a high-performance web server designed for
    performance experiments.
          
    This file is Copyright (C) 2004-2010  Tim Brecht

    Author: Tim Brecht <brecht@cs.uwaterloo.ca>.

    This code has evolved from code for servers by: 
       David Mosberger <david.mosberger@hp.com> and
       Abhishek Chandra <abhishek@cs.umass.edu>

    Contributions by Ashif Harji <asharji@cs.uwaterloo.ca>
    See AUTHORS file for list of additional contributors to the project.
  
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"
#include "options.h"
#include "info.h"
#include "myassert.h"
#include "state.h"
#include "debug.h"
#include "aio_layer.h"
#include "stats.h"
#include "syscalls.h"
#include "util.h"
#include "interest_set.h"
#include "fcgi.h"
#include "sock_listener.h"
#include "dll.h"

#include "seda.h"

/* #define LOCAL_DEBUG */
#include "local_debug.h"


infot *info;

#ifdef HAVE_AIO_LAYER
void *read_reg_addr = 0;
void *read_reg_handle = 0;
int   read_reg_size = 0;
void *write_reg_addr = 0;
void *write_reg_handle = 0;
int   write_reg_size = 0;
#endif /* HAVE_AIO_LAYER */

/* ---------------------------------------------------------------------- */
int
infot::info_init()
{
//    do_plock l(&info_init_lock);
#ifdef HAVE_AIO_LAYER
  int rc = 0;
#endif /* HAVE_AIO_LAYER */
  int i;
  //struct info *ip;
  int size;
  tmp_read = 0;
  tmp_reply = 0;
  app_t *app;
  appserver_t *appserver;

  /*
   * Compute the maximum number of info structs we could possibly use.
   * First, we'll need an info struct for each possible client connection.
   */
  num_info_structs = maxconns + 100;
//  num_info_structs = max_fds;

  /*
   * Finally, we'll need some extra info structs to accommodate any special
   * sockets which might use them.  Currently, this means one for each
   * FastCGI application server.
   */
  for (i = 0; i < options.app_count; i++) {
    app = &options.app[i];
    switch (app->req_type) {
      case REQ_FASTCGI:
	appserver = app->server;
	assert(appserver != NULL);
	do {
	  num_info_structs++;
	  appserver = appserver->next;
	} while (appserver != NULL);
	break;
      default:
	/* nothing to do for this type of application */
	break;
    }
  }

  printf("info_init: sizeof(struct req)  = %d mod cache_line_size = %d "
          "mod 4 = %d\n", 
          (int) sizeof(struct req),
          (int) sizeof(struct req) % cache_line_size,
          (int) sizeof(struct req) % 4);
  printf("info_init: sizeof(struct rep)  = %d mod cache_line_size = %d "
          "mod 4 = %d\n", 
          (int) sizeof(struct rep),
          (int) sizeof(struct rep) % cache_line_size,
          (int) sizeof(struct rep) % 4);

  printf("info_init: sizeof(struct info) = %d mod cache_line_size = %d "
          "mod 4 = %d\n", 
          (int) sizeof(struct info),
          (int) sizeof(struct info) % cache_line_size,
          (int) sizeof(struct info) % 4);

  size = num_info_structs * sizeof(struct info);
  if (size / (int) sizeof(struct info) != num_info_structs) {
    printf("info_init: overflow calculating size for info\n");
    exit(1);
  }

#if 1
  infomem = (struct info *) alloc_aligned(cache_line_size, size, 
         ALLOC_MALLOC, "info_init (info array):");

  printf("info_init: info entries = %d size = %lu\n", 
      num_info_structs, (unsigned long) size);
  // memset(info, 0, size);
#endif
  info = new struct info * [max_fds];

  printf("info_init: read buffer size = %d\n", options.read_buffer_size);
  size = num_info_structs * options.read_buffer_size;
  if (size / options.read_buffer_size != num_info_structs) {
    printf("info_init: overflow calculating size for tmp_read\n");
    exit(1);
  }

  /* NOTE: this is mmap'ed because it's an easy way to ensure
     that the data is page aligned, which IIRC is required for 
     aio_sock_mem_reg below
   */
  tmp_read = (char *) alloc_aligned(pagesize, size, ALLOC_MMAP, "info_init:");

  printf("info_init: read buffer entries = %d total size = %lu\n", 
      num_info_structs, (unsigned long) size);

  /* Depending on the underlying layer this may have to be done
   * for each chunk below - instead of all at once like this 
   */
  if (options.use_socket_aio) {
#ifdef HAVE_AIO_LAYER
    rc = aio_sock_mem_reg(tmp_read, size, &read_reg_handle);
    stats_reg_bytes(size);
    if (rc < 0) {
      printf("info_init: aio_sock_mem_reg failed (read buffer) "
	  "addr = %p size = %d handle = %p\n", tmp_read, size, read_reg_handle);
      exit(1);
    }

    PRINT_TIME(NOFD, &tnow, &tprev, "info_init: aio_sock_mem_reg start = %p "
      "size = %d handle = %p", tmp_read, size, read_reg_handle);
    read_reg_addr = tmp_read;
    read_reg_size = size;
#else
    printf("info_init: aio not handled\n");
    exit(1);
#endif /* HAVE_AIO_LAYER */
  }

  printf("info_init: reply buffer size = %d\n", options.reply_buffer_size);
  size = num_info_structs * options.reply_buffer_size;
  if (size / options.reply_buffer_size != num_info_structs) {
    printf("info_init: overflow calculating size for tmp_reply\n");
    exit(1);
  }

  tmp_reply = (char *) alloc_aligned(pagesize, size, ALLOC_MMAP,
      "info_init: (reply buffer)");
  printf("info_init: reply buffer entries = %d total size = %lu\n", 
      num_info_structs, (unsigned long) size);

  /* Depending on the underlying layer this may have to be done
   * for each chunk below - instead of all at once like this 
   */
  if (options.use_socket_aio) {
#ifdef HAVE_AIO_LAYER
    rc = aio_sock_mem_reg(tmp_reply, size, &write_reg_handle);
    stats_reg_bytes(size);
    if (rc < 0) {
      printf("info_init: aio_sock_mem_reg failed (reply buffer) "
        "addr = %p size = %d handle = %p\n", 
        tmp_reply, size, write_reg_handle);
      exit(1);
    }
    PRINT_TIME(NOFD, &tnow, &tprev, "info_init: aio_sock_mem_reg start = %p "
      "size = %d handle = %p", tmp_reply, size, write_reg_handle);

    write_reg_addr = tmp_reply;
    write_reg_size = size;
#else
    printf("info_init: aio not handled\n");
    exit(1);
#endif /* HAVE_AIO_LAYER */
  }

  for (i = 0; i < max_fds; i++) {
    info[i] = NULL;
    //info[i] = new struct info;
    //ip = info[i];
    //initInfoptr(ip);
#if 0
    ip->fsm_state = FSM_INITIAL_STATE;
    ip->sd = -1;
#ifdef HAVE_EPOLL_CTLV
    ip->sd_state = 0;
#endif /* HAVE_EPOLL_CTLV */
    ip->req.num = -1;
    ip->conn_num = -1;
    ip->req.static_buf = tmp_read;
    ip->req.dyn_buf = NULL;
    ip->req.buf = ip->req.static_buf;
    ip->req.buf_size = options.read_buffer_size;
    ip->req.end = ip->req.cur = ip->req.buf;
    PRINT_TIME(NOFD, &tnow, &tprev, "info_init: (read) req.buf[%d] = %p", 
       i, ip->req.buf);
    tmp_read += ip->req.buf_size;
    ip->rep.fd = -1;
    ip->rep.static_buf = tmp_reply;
    ip->rep.dyn_buf = NULL;
    ip->rep.buf = ip->rep.static_buf;
    ip->rep.buf_size = options.reply_buffer_size;
    ip->rep.end = ip->rep.cur = ip->rep.buf;
    tmp_reply += ip->rep.buf_size;
#ifdef QUEUE_BUFS
		ip->done_write = 0;
#endif
#ifdef HAVE_AIO_LAYER
    ip->req.read_reg_handle = read_reg_handle;
    ip->rep.write_reg_handle = write_reg_handle;
#endif /* HAVE_AIO_LAYER */
#endif // if 0
  } /* for */


  size = (max_fds + 1) * sizeof(int);
  if (size / (int) sizeof(int) != (max_fds + 1)) {
    printf("info_init: overflow calculating size for sd_map\n");
    exit(1);
  }

//  sd_map = (int *) alloc_aligned(cache_line_size, size, ALLOC_MALLOC,
//      "info_init: (sd_map)");
//  printf("info_init: sd_map buffer entries = %d total size = %lu\n", 
//      max_fds, (unsigned long) size);

//  for (i = 0; i <= max_fds; i++) {
//    sd_map[i] = NULL_INDEX;
//  }

  /* NOTE: I think that they must be done in this order */
  for (i = num_info_structs - 1; i >= 0; i--) {
    initInfoptr(&infomem[i]);
    info_free_entry(&infomem[i]);
  }

#ifdef LOCAL_DEBUG
  infoStack.info_free_print();
  infoStack.info_print();
  sd_map_free_print();
  sd_map_print();
#endif /* LOCAL_DEBUG */

  return 1;
} /* info_init */

void infot::initInfoptr( struct info *ip) {
    ip->fsm_state = FSM_INITIAL_STATE;
    ip->sd = -1;
#ifdef HAVE_EPOLL_CTLV
    ip->sd_state = 0;
#endif /* HAVE_EPOLL_CTLV */
    ip->req.num = -1;
    ip->conn_num = -1;
    ip->req.static_buf = tmp_read;
    ip->req.buf_size = options.read_buffer_size;
    tmp_read += ip->req.buf_size;
    ip->rep.static_buf = tmp_reply;
    ip->rep.buf_size = options.reply_buffer_size;
    tmp_reply += ip->rep.buf_size;
    ip->req.dyn_buf = NULL;
    ip->req.buf = ip->req.static_buf;
    ip->req.end = ip->req.cur = ip->req.buf;
    PRINT_TIME(NOFD, &tnow, &tprev, "info_init: (read) req.buf[%d] = %p", 
       i, ip->req.buf);
    ip->rep.fd = -1;
    ip->rep.dyn_buf = NULL;
    ip->rep.buf = ip->rep.static_buf;
     ip->rep.end = ip->rep.cur = ip->rep.buf;
#ifdef QUEUE_BUFS
		ip->done_write = 0;
#endif
#ifdef HAVE_AIO_LAYER
    ip->req.read_reg_handle = read_reg_handle;
    ip->rep.write_reg_handle = write_reg_handle;
#endif /* HAVE_AIO_LAYER */

}

void infot::clearInfoptr( int index ){
    if( info[index] != NULL ) {
#if 0
	delete (info[index]);
#else
	info_free_entry(info[index]);
#endif
	info[index] = NULL;
    } // if
}

/* ---------------------------------------------------------------------- */
void
infot::info_free_print(void)
{
  int i = info_first_free;

  printf("info_free_print: printing free info entries\n");
  while (i != NULL_INDEX) {
    printf("info[%d] is free\n", i);
    i = info[i]->next_free;
  }
}

/* ---------------------------------------------------------------------- */
void
infot::info_print(void)
{
  int i;

  printf("info_print: printing info entries\n");
  printf("info_print: info_first_free = %d\n", info_first_free);
  for (i = 0; i < num_info_structs; i++) {
    printf("info[%d].next_free = %d\n", i, info[i]->next_free);
  }
}

/* ---------------------------------------------------------------------- */
void
infot::info_print_status(void)
{
  int i;
  int sd_count = 0;

  printf("------------------------------------------------------------\n");
  printf("info_print_status: printing info status entries\n");
  printf("info_print_status: info_first_free = %d\n", info_first_free);
  printf("info_print_status: num_info_structs = %d\n", num_info_structs);
  for (i = 0; i < num_info_structs; i++) {
    if (info[i]->sd > 0) {
      sd_count++;
      printf("%6d type = %d sd = %6d sd_read_state = 0x%4x sd_write_state = 0x%4xfsm_state = %2d "
             "sock state = %c%c interest = %c%c conn_num = %6d cache = %12p HTTP 1.%d ", 
	      i, 
	      info[i]->type,
	      info[i]->sd, 
	      info[i]->sd_read_state,
	      info[i]->sd_write_state,
	      info[i]->fsm_state,
	      (is_sock_readable(info[i]->sd, 0) ? 'R' : '-'),
	      (is_sock_writable(info[i]->sd, 0) ? 'W' : '-'),
	      (interest_set_is_readable(info[i]->sd) ? 'R' : '-'),
	      (interest_set_is_writable(info[i]->sd) ? 'W' : '-'),
	      info[i]->conn_num,
	      info[i]->rep.cache, 
	      info[i]->req.close ? 0 : 1);
      if (info[i]->rep.cache) {
	printf("fd = %6d", info[i]->rep.cache->fd);
      }
      printf("\n");

    }
  }
  printf("sd_count = %d\n", sd_count);
}

/* ---------------------------------------------------------------------- */
struct info *
infot::info_add(int sd)
{
//  int index = NULL_INDEX;
  ldbg("info_add: entered sd = %d\n", sd);

  assert(sd >= 0);
  assert(sd <= max_fds);
//  assert(sd_map[sd] == NULL_INDEX);

//  index = infoStack.info_get_entry();

//  sd_map[sd] = sd;
  info[sd]->sd = sd;

  ldbg("info_add: returning sd = %d index = %d\n", sd, index);

  return(info[sd]);
}

/* ---------------------------------------------------------------------- */
struct info *
infot::info_reserve(void)
{
  int index = 0;
  struct info *ip;

  ldbg("info_reserve: entered\n");
//  index = info_get_entry();
  assert(index != NULL_INDEX);
  ip = info[index];

  ldbg("info_reserve: returning index = %d ip = %p\n", index, ip);
  return ip;
}

/* ---------------------------------------------------------------------- */
void
infot::info_unreserve(struct info *ip)
{
  int index;

  ldbg("info_register: entered ip = %p\n", ip);
  assert(ip != NULL);
  index = ip - *info;
  assert(index >= 0);
  assert(index <= max_sock_sd);
  assert(ip == info[index]);

  info_free_entry(info[index]);
}

/* ---------------------------------------------------------------------- */
void
infot::info_register(struct info *ip, int sd)
{
  int index;

  ldbg("info_register: entered ip = %p sd = %d\n", ip, sd);
  assert(ip != NULL);
  assert(sd >= 0);
  assert(sd <= max_fds);
  index = ip - *info;
  assert(index >= 0);
  assert(index <= max_sock_sd);
  assert(ip == info[index]);

//  sd_map[sd] = index;
  info[index]->sd = sd;
  ldbg("info_register: returning index = %d\n", index);
}

/* ---------------------------------------------------------------------- */
void
infot::info_delete(int sd)
{
  int index = NULL_INDEX;
  assert(sd >= 0);
  assert(sd <= max_fds);
  ldbg("info_delete: entered sd = %d sd_map_first_free = %d\n",
      sd, info_first_free);

  info[sd]->sd = -1;
  info[sd]->fsm_state = FSM_INITIAL_STATE;

  index = sd;
  assert(index != NULL_INDEX);
  info[index]->sd = -1;
//  sd_map[sd] = NULL_INDEX;
//  infoStack.info_free_entry(index);

//  printf("info_del:%d, task: %p, index: %d, %p, info_last_free: %d\n",sd, &uThisTask(), index, &info[index], info_last_free);
  ldbg("info_delete: returning sd = %d sd_map_first_free = %d\n",
      sd, sd_map_first_free);

}

/* ---------------------------------------------------------------------- */
void
infot::info_free_entry(struct info *ip)
{
#if 0

  int old_first = info_first_free;
  int old_next = NULL_INDEX;

  assert((index >= 0) && (index <= max_sock_sd));

  if (info_first_free != NULL_INDEX) {
    old_next = info[info_first_free].next_free;
  }

  ldbg("info_free_entry: entered index = %d old_first = %d old_next = %d\n",
	index, old_first, old_next);

  info_first_free = index;
  info[info_first_free].next_free = old_first;
  ldbg("info_free_entry: freed index = %d first_free = %d\n",
	index, info_first_free);
  if (info_first_free != NULL_INDEX) {
     ldbg("info_free_entry: next = %d\n", info[info_first_free].next_free);
  }

#else

//  int old_last = info_last_free;

  int index = ip - infomem;
  do_plock l(&info_lock);
  assert((index >= 0) && (index <= max_sock_sd));
  
  if ( info_last_free != NULL_INDEX) {
    infomem[info_last_free].next_free = index;
  }

  info_last_free = index;
  infomem[info_last_free].next_free = NULL_INDEX;

  if ( info_first_free == NULL_INDEX ) {
      info_first_free = index;
  } // if

//  printf("info_free:first_free: %d, index: %d, %p, info_last_free: %d\n",info_first_free, index, &info[index], info_last_free);
//  printf("info_free: info_last_free:%d, old_last:%d, index:%d\n", info_last_free, old_last, index);

#endif
  
}

/* ---------------------------------------------------------------------- */
struct info *
infot::info_get_entry(void)
{
  do_plock l(&info_lock);
  int index = info_first_free;
#ifdef LOCAL_DEBUG
  ldbg("info_get_entry: first_free = %d next = %d cur_conns = %d\n",
	info_first_free, infomem[info_first_free].next_free, cur_conns);
#endif /* LOCAL_DEBUG */

#ifdef LOCAL_DEBUG
  info_free_print();
  info_print();
  sd_map_free_print();
  sd_map_print();
#endif /* LOCAL_DEBUG */

  assert(info_first_free != NULL_INDEX);
  assert(info_first_free >= 0); 
  assert(info_first_free <= max_fds);

  info_first_free = infomem[index].next_free;
  infomem[index].next_free = NULL_INDEX;

  if( info_last_free == index ) {
      info_last_free = NULL_INDEX;
  } // if

  ldbg("info_get_entry: returning index = %d first_free = %d\n",
	index, info_first_free);
  
//  printf("info_get_entry: returning index = %d first_free = %d, last_free: %d, num_idle: %d\n",index, info_first_free, info_last_free, get_num_idle());

  return(&infomem[index]);
}

#if 0
/* ---------------------------------------------------------------------- */
void
infot::sd_map_print(void)
{
  int i;
  printf("sd_map_print:\n");

  for (i = 0; i <= max_fds; i++) {
    printf("sd_map[%d] = %d\n", i, sd_map[i]);
  }
}

/* ---------------------------------------------------------------------- */
void
infot::sd_map_free_print(void)
{
  int tmp = sd_map_first_free;
  printf("sd_map_free_print\n");

  while (tmp != NULL_INDEX) {
    printf("free = %d\n", tmp);
    tmp = sd_map[tmp];
  }
}
#endif
/* ---------------------------------------------------------------------- */
struct info *
infot::info_ptr(int sd)
{
  int index;
  struct info *ip;

  index = info[sd]->sd;
//  index = info_index(sd);
  if (index == NULL_INDEX) {
    ip = NULL;
  } else {
    ip = info[index];
    assert(ip->sd == sd);
  }
  ldbg("info_ptr: sd = %d index = %d ip = %p\n", sd, index, ip);
  return(ip);
}

/* ---------------------------------------------------------------------- */
struct info *
infot::info_ptr_no_assert(int sd)
{
  int index;
  struct info *ip;

  index = sd;
//  index = info_index(sd);
  if (index == NULL_INDEX) {
    ip = NULL;
  } else {
    ip = info[index];
  }
  if( ip == NULL) {
//      info[index] = new struct info;
#if 1
      info[index] = info_get_entry();
#else
      info[index] = (struct info *) malloc( sizeof(struct info) + options.read_buffer_size + options.reply_buffer_size);
      ::info->initInfoptr(info[index]);
#endif
      ip = info[index];
  }
  ldbg("info_ptr: sd = %d index = %d ip = %p\n", sd, index, ip);
  return(ip);
}

/* ---------------------------------------------------------------------- */
struct req *
info_req_ptr(int sd)
{
  struct info *ip = info_ptr(sd);
  assert(ip);
  return(&(ip->req));
}

/* ---------------------------------------------------------------------- */
struct rep *
info_rep_ptr(int sd)
{
  struct info *ip = info_ptr(sd);
  assert(ip);
  return(&(ip->rep));
}

/* ---------------------------------------------------------------------- */
int
info_get_fsm_state(int sd)
{
  struct info *ip = info_ptr(sd);
  assert(ip);
  return(ip->fsm_state);
}

/* ---------------------------------------------------------------------- */
void
info_set_fsm_state(int sd, enum info_fsm_state state)
{
  struct info *ip = info_ptr(sd);
  assert(ip);
  set_fsm_state(ip, state);
}

/* ---------------------------------------------------------------------- */
int
infot::info_index(int sd)
{
  int index = NULL_INDEX;
  ldbg("info_index: looking for sd = %d\n", sd);
  assert(sd >= 0);
  assert(sd <= max_fds);

  index = sd;
  assert(index <= max_sock_sd);

  ldbg("info_index: sd = %d returning %d\n", sd, index);
  return(index);
}

/* ---------------------------------------------------------------------- */
int
infot::info_map_valid(int sd)
{
  int valid = 0;
  ldbg("info_map_done: looking for sd = %d\n", sd);
  assert(sd >= 0);
  assert(sd <= max_fds);

  if (sd == NULL_INDEX) {
    valid = 0;
  } else {
    valid = 1;
  }

  ldbg("info_map_valid: sd = %d returning %d\n", sd, valid);
  return(valid);
}

/* ---------------------------------------------------------------------- */
/* Close the fd used for replies.  */
int
info_close_rep_fd(struct info *ip)
{
  int peer_sd;
  struct info *peer_ip;
  int rc = 0;

  PRINT_TIME(ip->sd, &tnow, &tprev, "info_close_rep_fd: ip->type = %d fd = %d",
      ip->type, ip->rep.fd);
  PRINT_TIME(ip->sd, &tnow, &tprev, "info_close_rep_fd: ip->req.type = %d ip->rep.use_cache = %d",
      ip->req.type, ip->rep.use_cache);

  if (ip->rep.fd < 0) {
    return 0;
  }

  /* we shouldn't be closing stdin, stdout, or stderr */
  assert(ip->rep.fd >= 3);

  switch (ip->type) {
    case INFO_CLIENT:
      switch (ip->req.type) {
	case REQ_STATIC:
	  if (!ip->rep.use_cache) {
	    /* only done if not using the cached copy */
	    rc = close_file(ip->rep.fd);
	  }
	  break;
	case REQ_FASTCGI:
	  peer_sd = ip->rep.fd;
	  peer_ip = info_ptr(peer_sd);
	  assert(peer_ip != NULL);
	  assert(peer_ip->rep.fd == ip->sd);
	  /* the FastCGI app is free to service another request */
          PRINT_TIME(ip->sd, &tnow, &tprev, "info_close_rep_fd: peer_sd = %d", peer_sd);
	  fcgi_service_next_request(peer_ip);
	  break;
#ifdef SPECWEB99_FAKE
	case REQ_SPECWEB99:
	  /* nothing to do */
	  break;
#endif
	default:
	  printf("info_close_rep_fd(): invalid ip->req.type %d\n",
	      ip->req.type);
	  assert(0);
      }
      break;

    default:
      printf("info_close_rep_fd(): invalid ip->type %d\n", ip->type);
      assert(0);
  }

  /* invalidate the fd */
  ip->rep.fd = -1;

  return rc;
}

/* ---------------------------------------------------------------------- */
/*
 * Use this function to reset rep.buf to point to rep.static_buf, and to
 * reset rep.cur and rep.end so that there's no data in the buf.
 * IMPORTANT: Do not use this function as part of initialization of an
 * info struct; fields in the info struct are assumed to be meaningful.
 */
void
info_reset_rep_buf(struct info *ip)
{
  app_t *app;
  int sd;
#ifdef AVOID_BUF_REUSE
  int ret=0;
  struct info *fcgi_ip, *first_ip=0;
#endif
  /*
   * If we had allocated a dynamic buffer for this reply, we should free
   * it now.
   */
  if (ip->rep.dyn_buf != NULL) {
    //printf("non-null buffer, cleanign\n");
#ifdef QUEUE_BUFS
		remove_from_list(dyn_bufpool->used_buf_q, ip->rep.buf_dll);
    ip->rep.buf_dll = NULL;
#endif

	bufpool_free(dyn_bufpool, ip->rep.dyn_buf);
    ip->rep.dyn_buf = NULL;

    /*
     * If this is now the only free dynamic buffer, check to see if there
     * are any applications that are blocked, waiting for a free buffer.
     */
    //if (bufpool_num_free(dyn_bufpool) == 1) {
      app = app_unblock_next();
	//printf("unblocking a blocked app-server\n");
      if (app != NULL) {
	switch (app->req_type) {
	  case REQ_FASTCGI:
	    /*
	     * Recover the request that caused the application to block
	     * in the first place (there *must* be one), and process it
	     * just as if it were arriving now.
	     */
	    sd = queue_get(app->req_q);
	    assert(sd >= 0);
	    fcgi_process_request(sd);
	    break;
	  default:
	    printf("info_reset_rep_buf(): invalid app->req_type %d\n",
		app->req_type);
	    assert(0);
	}
      }
#ifdef AVOID_BUF_REUSE
      else {
	//printf("no blocked app servers\n");
	//while (ret != 1) {
        fcgi_ip = fcgi_find_idle(0);
	while(fcgi_ip != first_ip) {
		if (fcgi_ip != NULL) {
			if (first_ip != 0) {
				if (first_ip != fcgi_ip)
					ret = fcgi_service_next_request(fcgi_ip);
				else
					break;
			}
			else {
				first_ip = fcgi_ip;
				ret = fcgi_service_next_request(fcgi_ip);
			}
				
		}
		else
			break;
	}
       }		
#endif
      }
  //}
  //else
//	printf("nothing to do\n");

#ifdef QUEUE_BUFS
	ip->done_write = 0;
#endif

  /*
   * If we had switched the active reply buffer to something other than our
   * static buffer, switch it back now.
   */
  if (ip->rep.buf != ip->rep.static_buf) {
    assert(ip->type == INFO_CLIENT);
    ip->rep.buf = ip->rep.static_buf;
    ip->rep.buf_size = options.reply_buffer_size;
  }

  /*
   * Finally, reset our 'cur' and 'end' pointers to flush any stale data.
   */
  ip->rep.end = ip->rep.cur = ip->rep.buf;
}

/* ---------------------------------------------------------------------- */
const char *
info_req_state_str(enum info_req_state state)
{
  switch (state)
  {
    case PARSE_METHOD:
      return "PARSE_METHOD";
    case PARSE_URI:
      return "PARSE_URI";
    case PARSE_VERSION:
      return "PARSE_VERSION";
    case PARSE_SKIP:
      return "PARSE_SKIP";
    case PARSE_CR_1:
      return "PARSE_CR_1";
    case PARSE_LF_1:
      return "PARSE_LF_1";
    case PARSE_HEADER_NAME:
      return "PARSE_HEADER_NAME";
    case PARSE_COLON:
      return "PARSE_COLON";
    case PARSE_COOKIE:
      return "PARSE_COOKIE";
    case PARSE_IF_MODIFIED_SINCE:
      return "PARSE_IF_MODIFIED_SINCE";
    case PARSE_CONTENT_LENGTH:
      return "PARSE_CONTENT_LENGTH";
    case PARSE_CR_2:
      return "PARSE_CR_2";
    case PARSE_LF_2:
      return "PARSE_LF_2";
    case PARSE_DONE:
      return "PARSE_DONE";
    /* FASTCGI states */
    case FCGI_READING_HEADER:
      return "FCGI_READING_HEADER";
    case FCGI_READING_END_REQUEST:
      return "FCGI_READING_END_REQUEST";
    case FCGI_READING_STDOUT:
      return "FCGI_READING_STDOUT";
    case FCGI_READING_STDERR:
      return "FCGI_READING_STDERR";
    case FCGI_READING_GET_VALUES_RESULT:
      return "FCGI_READING_GET_VALUES_RESULT";
    case FCGI_READING_UNKNOWN_TYPE:
      return "FCGI_READING_UNKNOWN_TYPE";
#ifdef SPECWEB99_FAKE
    /* SPECweb99 states */
    case SPECWEB99_REQUEST_DONE:
      return "SPECWEB99_REQUEST_DONE";
    case SPECWEB99_RESPONSE_DONE:
      return "SPECWEB99_RESPONSE_DONE";
    case SPECWEB99_SOCKET_CLOSED:
      return "SPECWEB99_SOCKET_CLOSED";
#endif
  }

  return "unknown";
}

/* ---------------------------------------------------------------------- */
const char *
info_req_method_str(enum info_req_method method)
{
  switch (method)
  {
    case HTTP_NONE:
      return "NONE";
    case HTTP_SIMPLE:
      return "SIMPLE";
    case HTTP_GET:
      return "GET";
    case HTTP_HEAD:
      return "HEAD";
    case HTTP_POST:
      return "POST";
    case HTTP_NOT_IMPL:
      return "NOT_IMPL";
    case HTTP_RETURN_ERROR:
      return "RETURN_ERROR";
  }

  return "unknown";
}

/* ---------------------------------------------------------------------- */
void
infot::store_sock_addr(int sd, const sockaddr_t *addr)
{
  struct info *ip;
  struct info_listen_struct *ilp;

  if (sock_is_listener(sd)) {
    ilp = &info_listener[sd];
    memcpy(&ilp->addr, addr, sizeof(sockaddr_t));
  } else {
    ip = info_ptr(sd);
    assert(ip != NULL);
    memcpy(&ip->addr, addr, sizeof(sockaddr_t));
  }
}

/* ---------------------------------------------------------------------- */
void
infot::get_sock_addr(int sd)
{
  struct info *ip;

  ip = info_ptr(sd);
  assert(ip != NULL);

  ip->addr.len = sizeof(ip->addr.u);
  if (getsockname(sd, &ip->addr.u.sa, &ip->addr.len) < 0) {
    printf("get_sock_addr(): failed\n");
    memset(&ip->addr, '\0', sizeof(ip->addr));
  } else {
    switch (ip->addr.u.sa.sa_family) {
      case AF_INET:
	ldbg("get_sock_addr(): AF_INET addr %s port %d\n",
	    inet_ntoa(ip->addr.u.sa_in.sin_addr), ip->addr.u.sa_in.sin_port);
	break;
      case AF_UNIX:
	ldbg("get_sock_addr(): AF_UNIX path %s\n",
	    ip->addr.u.sa_un.sun_path);
	break;
      default:
	printf("get_sock_addr(): unsupported family [%d]\n",
	    ip->addr.u.sa.sa_family);
	memset(&ip->addr, '\0', sizeof(ip->addr));
	break;
    }
  }
}


int info_index(int sd) {
    return info->info_index(sd);
}

struct info *info_ptr(int sd) {
    return info->info_ptr(sd);
}

struct info *info_ptr_no_assert(int sd){
    return info->info_ptr_no_assert(sd);
}

void info_delete(int sd){
    info->info_delete(sd);
}

int info_map_valid(int sd){
    return info->info_map_valid(sd);
}

void store_sock_addr(int sd, const sockaddr_t *addr){
    return info->store_sock_addr(sd, addr);
}

struct info *info_add(int sd){
    return info->info_add(sd);
}

void get_sock_addr(int sd){
    info->get_sock_addr(sd);
}

void info_print_status(void){
    info->info_print_status();
}


