/*
    userver -- (pronounced you-server or micro-server).
    This file is part of the userver, a high-performance web server designed for
    performance experiments.
          
    This file is Copyright (C) 2004-2010  Tim Brecht

    Author: Tim Brecht <brecht@cs.uwaterloo.ca>.

    This code has evolved from code for servers by: 
       David Mosberger <david.mosberger@hp.com> and
       Abhishek Chandra <abhishek@cs.umass.edu>

    See AUTHORS file for list of contributors to the project.
  
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/


#ifndef INFO_H
#define INFO_H

#include <limits.h>
#include <time.h>

#include "cache.h"
#include "util.h"
#include "app.h"
#include "dll.h"

/* NOTE: I'd like to try to keep these sizes defined so that
 * the size of a struct info is a multiple of the cache line size
 * while also trying to keep them small enough to keep
 * the total bytes consumed for the info array reasonable.
 * Note that we may allocate 15000 - 30000 struct info's.
 */
#ifdef IA64
#define MAX_TMP_LEN		   (64)
#define MAX_URI_LEN		   (64+48+12)  /* part of req */
#define MAX_COOKIE_LEN		   (32+8+30)   /* part of req */
#define MAX_SPECWEB99_OBJ_NAME_LEN (32+28+2)   /* part of rep */
#else
#define MAX_TMP_LEN		   (64)
#define MAX_URI_LEN		   (64+64+12)  /* part of req */
#define MAX_COOKIE_LEN		   (32+8+62)   /* part of req */
#define MAX_SPECWEB99_OBJ_NAME_LEN (32+28+2)   /* part of rep */
#endif

#define MAX_CONTENT_LEN		(INT_MAX)
#define NULL_INDEX		(-1)

//extern int *sd_map;			/* mapping from sd to info index */
//extern struct info *info;		/* info about each connection */

enum info_req_state {
  /* STATIC states */
  PARSE_METHOD = 0,
  PARSE_URI,
  PARSE_VERSION,
  PARSE_SKIP,
  PARSE_CR_1,
  PARSE_LF_1,
  PARSE_HEADER_NAME,
  PARSE_COLON,
  PARSE_COOKIE,
  PARSE_IF_MODIFIED_SINCE,
  PARSE_CONTENT_LENGTH,
  PARSE_CR_2,
  PARSE_LF_2,
  PARSE_DONE,
  /* FASTCGI states */
  FCGI_READING_HEADER,
  FCGI_READING_END_REQUEST,
  FCGI_READING_STDOUT,
  FCGI_READING_STDERR,
  FCGI_READING_GET_VALUES_RESULT,
  FCGI_READING_UNKNOWN_TYPE,
#ifdef SPECWEB99_FAKE
  /* SPECweb99 states */
  SPECWEB99_REQUEST_DONE,
  SPECWEB99_RESPONSE_DONE,
  SPECWEB99_SOCKET_CLOSED
#endif
};

enum info_req_method {
  HTTP_NONE,
  HTTP_SIMPLE,
  HTTP_GET,
  HTTP_HEAD,
  HTTP_POST,
  HTTP_NOT_IMPL,
  HTTP_RETURN_ERROR
};

struct req {
  int num;
  req_type_t type;
  app_t *app;
  enum info_req_state state;
  int skip_lws;			/* flag: skip LWS to get to next token? */
  enum info_req_method method;
  /*
   * We essentially use "close" as an HTTP version flag.  A value of 0 means
   * HTTP/1.1, while a value of 1 means HTTP/1.0.  If a new version of HTTP
   * comes out and we want to support it, then maybe we'll store a proper
   * HTTP version number.
   */
  int close;
  char *query_string;		/* points into uri[], or NULL if n/a */
  int cookie_len;
  int content_len;
  int padding_len;
  char *static_buf;		/* pre-allocated buffer */
  char *dyn_buf;		/* dynamically-allocated buffer */
  char *buf;			/* current buffer (e.g. one of the above) */
  int buf_size;		        /* size of current buffer */
  char *cur;			/* beginning of data in current buffer */
  char *end;			/* next spot for data in current buffer */
// #ifdef HAVE_AIO_LAYER
  void *read_reg_handle;
// #endif
  int tmp_len;
  /*
   * The tmp[] array is used for building up a single token (the HTTP
   * method, the HTTP version, or a header name).  It must be large enough
   * to hold the longest _supported_ header name (plus NUL terminator).
   * The tmp_len variable refers to the amount of data currently in tmp[].
   */
  char tmp[MAX_TMP_LEN];
  char cookie[MAX_COOKIE_LEN];
  char uri[MAX_URI_LEN];
  time_t ifmodsince;		/* from If-Modified-Since header */
};

struct rep {
  int fd;			/* associated file or socket descriptor */

  cacheinfo *cache;
  char *static_buf;		/* pre-allocated buffer */
  char *dyn_buf;		/* dynamically-allocated buffer */
  char *buf;			/* current buffer (e.g. one of the above) */
  int buf_size;		        /* size of current buffer */
  char *cur;			/* beginning of data in current buffer */
  char *end;			/* next spot for data in current buffer */
// #ifdef HAVE_AIO_LAYER
  void *write_reg_handle;
// #endif
  int total_len;
  int nwritten;
  int offset;
  struct iovec iovec[2];     /* iovec[0] for header iovec[1] for file */
  char use_cache;   /* if set no need to open, read from or close file */
  char is_corked;
  int http_status;              /* HTTP status code of the reply (e.g. 200, 404, etc.) */
#ifdef SPECWEB99_FAKE
  char objectname[MAX_SPECWEB99_OBJ_NAME_LEN];
  int objectlen;
#endif

#ifdef QUEUE_BUFS
	struct dll_element *buf_dll;
#endif

#ifdef IA64
  char __pad[8];
#endif
};

enum info_type {
  INFO_CLIENT = 0,
  INFO_LISTENER,
  INFO_FASTCGI
};

enum info_fsm_state {
  FSM_UNKNOWN = 0,
  /* client states */
  FSM_INITIAL_STATE,
  FSM_CONNECTING,
  FSM_READING_REQUEST,
  FSM_READ_ARRIVED_WHILE_WRITING,
  FSM_WRITING_REPLY,
  FSM_READING_NEXT_REQUEST,
  FSM_CLOSING,
  FSM_CLOSED,
  FSM_LISTENING,
  /* FastCGI request states */
  FSM_FCGI_WRITING_SET_VALUES,
  FSM_FCGI_WRITING_BEGIN_REQUEST,
  FSM_FCGI_WRITING_PARAMS_1,
  FSM_FCGI_WRITING_PARAMS_2,
  FSM_FCGI_WRITING_END_PARAMS,
  FSM_FCGI_WRITING_STDIN,
  FSM_FCGI_WRITING_END_STDIN,
  FSM_FCGI_WAITING_FOR_REPLY,
  FSM_CONN_ERROR,
  FSM_STATE_MAX
};

struct info {
  int sd;
  sockaddr_t addr;
  enum info_type type;
	
#ifdef QUEUE_BUFS
	int done_write;
#endif

  // TODO
  // All fields below this point are dependent upon type (INFO_CLIENT, etc.).
  // It would make sense to just have a void *ptr that pointed to this
  // type-specific data.

  app_t *app;
  appserver_t *appserver;
  int parent_sd;        /* socket that accepted this connection */
  int conn_num;
  int next_free;
// To try to get sizing figured out whether AIO is being used or not.
// #ifdef HAVE_AIO_LAYER
  int save_result;      /* save out of order aio completion info */
  int save_err;
// #endif /* HAVE_AIO_LAYER */

  unsigned int sd_read_state;
  unsigned int sd_write_state;

  /* NOTE: if you change this be sure to change the fsm_string in state.c */
  enum info_fsm_state fsm_state;

#ifdef SPECWEB99_FAKE
  void *priv;		/* private handle to a buffer - same as TUX */
#endif

  struct rep rep;
  struct req req;

  char cbuf[0];
};

struct admin_info {
    int affinity_queue;
    int subnet_id;
};

class infot {
//    int sd_map_first_free;
    int info_first_free;
    int info_last_free;
//    int *sd_map;
    int num_info_structs;
    struct info **info;
    struct info *infomem;
    char *tmp_read;
    char *tmp_reply;
    pthread_mutex_t info_lock;
  public:

    infot(){
//	sd_map_first_free = NULL_INDEX;
	info_first_free = NULL_INDEX;
	info_last_free = NULL_INDEX;
	num_info_structs = 0;
	pthread_mutex_init(&info_lock, NULL);
	info_init();
    }
    
    struct info *info_add(int sd);
    void info_delete(int sd);
    int info_get_fsm_state(int sd);
    void info_set_fsm_state(int sd, enum info_fsm_state state);
    struct info *info_ptr(int sd);
    struct info *info_ptr_no_assert(int sd);
    struct req *info_req_ptr(int sd);
    struct rep *info_rep_ptr(int sd);
    int info_index(int sd);
    int info_init(void);
    void info_print_status(void);
    struct info *info_reserve(void);
    void info_unreserve(struct info *ip);
    void info_register(struct info *ip, int sd);
    int info_map_valid(int sd);
    const char *info_req_state_str(enum info_req_state state);
    const char *info_req_method_str(enum info_req_method method);
    void store_sock_addr(int sd, const sockaddr_t *addr);
    void get_sock_addr(int sd);
//    void sd_map_free_print(void);
//    void sd_map_print();
    
    void info_free_entry(struct info *ip);
    struct info *info_get_entry();
    void info_free_print();
    void info_print();

    void initInfoptr( struct info *ip);
    void clearInfoptr( int index );
};

extern int info_index(int sd);
extern struct info *info_ptr(int sd);
extern struct info *info_ptr_no_assert(int sd);
extern void info_reset_rep_buf(struct info *ip);
extern void info_delete(int sd);
extern int info_close_rep_fd(struct info *ip);
extern int info_map_valid(int sd);
extern void store_sock_addr(int sd, const sockaddr_t *addr);
extern struct info *info_add(int sd);
extern void get_sock_addr(int sd);
extern void info_print_status(void);

#endif /* INFO_H */
