/*
    userver -- (pronounced you-server or micro-server).
    This file is part of the userver, a high-performance web server designed for
    performance experiments.
          
    This file is Copyright (C) 2005  David Pariag

    Authors: David Pariag <db2paria@cs.uwaterloo.ca>
    See AUTHORS file for list of contributors to the project.
  
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/


/*
 * Title          : http_log_gen.c
 * Author         : David Pariag
 * Created        : Feb 4th 2005
 * Last Modified  : Nov 23rd 2005
 *
 */

/********************************************************************
 * The following means are based on the Pareto and Weibull 
 * distributions that are discussed in the SURGE paper 
 * (Barford and Crovella)
 ********************************************************************/
#define WEIBULL_MEAN    (0.343)
#define PARETO_MEAN  (3.00)

#include "http_log_gen.h"
#include "session.h"
#include "distributions.h"

opts genOpts;

session_list sessions[MAX_SESSION];

static struct option long_opts[] = {
  {"zip", required_argument, 0, 1},
  {"delta", required_argument, 0, 2},
  {"timer-esc", no_argument, 0, 3},
  {"prime-cache", no_argument, 0, 4},
  {"clear-stats", no_argument, 0, 5},
  {"no-http-1.0", no_argument, 0, 6},
  {"help", no_argument, 0, 7 },
  {"num-dirs", required_argument, 0, 8},
  {"min-sessions", required_argument, 0, 9},
  {"max-sessions", required_argument, 0, 10},
  {"fileset-size", required_argument, 0, 11}, 
  {"tolerance", required_argument, 0, 12},
  {"use-think-times", no_argument, 0, 13},
  {"random-seed", required_argument, 0, 14},
  {"num-logs", required_argument, 0, 15},
  {"fixed-think-times", required_argument, 0, 16},
  {"dynamic-only", no_argument, 0, 17},
  {"root-dir", required_argument, 0, 18},
  {"dynamic-uri", required_argument, 0, 19},
  {"static-only", no_argument, 0, 20},
  {"num-clients", required_argument, 0, 21},
  {0,0,0,0}
};



void usage(char * progname)
{
  printf("\nUsage: %s [options]. Available options are:\n", progname);
  printf("  --fileset-size   #      Limit the size of the file set to # MB (just below)\n");
  printf("  --dynamic-only          Only generate dynamic content requests\n");
  printf("  --static-only           Only generate static content requests\n");
  printf("  --dynamic-uri <uri>     Dynamic uri/script that will be used to generate dynamic requests\n");
  printf("  --num-dirs       #      Number of directories in workload (numbered from zero)\n");
  printf("  --num-clients    #      Number of SPECweb99 clients to simulate (used for ad rotation)\n");
  printf("  --num-logs       #      Specify the number of log files to generate\n");
  printf("  --tolerance      x%s     Run until distribution tolerances are within a factor of x%s\n", "%", "%");
  printf("  --delta          y%s     Make initial tolerance stricter by y%s to compensate for discarded sessions\n", "%", "%");
  printf("  --min-sessions   #      Minimum number (#) of sessions\n");
  printf("  --max-sesions    #      Maximum number (#) of sessions\n");
  printf("  --use-think-times       Use non-zero think times\n"); 
  printf("  --fixed-think-times #   Use # milliseconds for all think times\n");
  printf("  --root-dir  <dir>       Specify the directory that contains the SPECweb99 fileset\n");
  printf("  --random-seed    #      Specify the random seed\n");
  printf("  --zip       <zipfile>   Use tar and gzip to create an archive containing all log files\n");
  printf("  --timer-esc             Generate escape sequences to turn timers on and off\n");
  printf("  --prime-cache           Produce a log that can be used to prime the cache\n");
  printf("  --clear-stats           Add special userver url to clear stats\n");
  printf("  --help                  Print this message\n");   
#if 0
  printf("-l #   : length each session must be # (NOTE: not specweb-like)\n");
#endif
  exit(1);
}


/* Set default values for most command line options */
void initOpts () {
  genOpts.dynamic_only = 0;
  genOpts.static_only = 0;
  genOpts.clear_stats = 0;
  genOpts.num_dirs = 1;
  genOpts.num_clients = 1000;
  genOpts.timer_esc = 0;
  genOpts.prime_cache = 0;
  genOpts.min_sessions = 1;
  genOpts.max_sessions = 1000000;
  genOpts.fileset_size = -1;
  genOpts.tolerance = 0.1;
  genOpts.delta = 0.01;
  genOpts.use_think_times = 0;
  genOpts.no_http_10 = 0;
  genOpts.random_seed = 0;
  genOpts.num_logs = 1;
  genOpts.zip = 0;
  genOpts.fixed_think = -1.0;
  strncpy(genOpts.root_dir, REQUESTDIR, MAX_DIR_LEN);
  strncpy(genOpts.dynamic_uri, DYNAMIC_SCRIPT, MAX_URI);
}

void checkOpts() {
  if( genOpts.prime_cache || genOpts.timer_esc ) {
    printf("Neither --prime-cache nor --timer-esc is currently supported\n");
    exit(1);
  }

}


/*****************************************************************
 * Process command line args. Set flags in genOpts global to
 * reflect the processed args.
 *****************************************************************/
void process_args(int argc, char * argv[] ) {
  int c;
  int long_index;
  int start = 0;

  while ((c = getopt_long(argc, argv, "abcdefghijklmnopqrstuvwxyz", long_opts, &long_index )) != EOF) {
    switch(c) {      
    case 1:
      genOpts.zip = 1;
      strcpy(genOpts.zip_file, optarg);
      break;
    case 2:
      genOpts.delta = atof(optarg) / 100.0;
      break;      
    case 3:
      printf("Generating escape sequences to turn timers on and off\n");
      genOpts.timer_esc = 1;
      break;
    case 4:
      printf("Generating requests that will prime the cache\n");
      genOpts.prime_cache = 1;      
      break;
    case 5:
      printf("Generating requests to clear server stats\n");
      genOpts.clear_stats = 1;
      break;
    case 6:
      printf("Generating HTTP 1.1 requests only\n");
      genOpts.no_http_10 = 1;
      break;
    case 7:
      usage(argv[0]);
      break;      
    case 8:
      genOpts.num_dirs = atoi(optarg);
      printf("Limiting directories to a maximum of %d\n", genOpts.num_dirs);
      if (genOpts.fileset_size != -1) {
	printf("Can't specify both maximum directory and maximum size\n");
	exit(1);
      }
      break;
    case 9:
      genOpts.min_sessions = atoi(optarg);
      printf("Generating a minimum of %d sessions\n", genOpts.min_sessions);
      if (genOpts.prime_cache) {
	printf("Can't prime the cache and specify "
	       "the number of sessions\n");
	exit(1);
      }
      break;
      
    case 10:      
      genOpts.max_sessions = atoi(optarg);
      printf("Generating a maximum of %d sessions\n", genOpts.max_sessions);
      if (genOpts.prime_cache) {
	printf("Can't prime the cache and specify "
	       "the number of sessions\n");
	exit(1);
      }
      break;
            
    case 11:
      genOpts.fileset_size = atoi(optarg);
      printf("Maximum fileset size is %d MB => ", genOpts.fileset_size);
      genOpts.num_dirs = (int) floor((double) ((double)genOpts.fileset_size * 1024.0 * 1024.0) / 
			     (double) (BYTES_PER_DIR));
      printf("maximum of %d directories...\n", genOpts.num_dirs);
      printf("Using %d directories gives fileset size of %8.4f = %.2lf MB\n", 
             genOpts.num_dirs, ((double)genOpts.num_dirs * (double)BYTES_PER_DIR),
             (double) ((double)genOpts.num_dirs * (double)BYTES_PER_DIR) / 1024.0 / 1024.0);
      break;
      
    case 12:
      genOpts.tolerance = atof(optarg) / 100.0;
      printf("Ensuring that distributions are within a factor of %1.1f\n", genOpts.tolerance);
      break;
      
    case 13:
      genOpts.use_think_times = 1;
      break;
      
    case 14:
      genOpts.random_seed = (unsigned int) atoi(optarg);
      break;      
      
    case 15:
      genOpts.num_logs = atoi(optarg);
      break;

    case 16:
      genOpts.fixed_think = (atof(optarg) / 1000.0);
      break;

    case 17:
      genOpts.dynamic_only = 1;
      break;

    case 18:
	/* Make sure the root dir has a leading slash */
      if( strlen(optarg) > 0 ) {
	if(optarg[0] != '/' ) {
	  genOpts.root_dir[0] = '/';
	  start++;
	}
	strncpy((genOpts.root_dir + start), optarg, MAX_DIR_LEN - 2);
	
	/* Make sure the root dir has a trailing slash */
	if(optarg[strlen(optarg)-1] != '/' ) {
	  genOpts.root_dir[strlen(genOpts.root_dir)] = '/';
	}
      }
	break;
    case 19:
      strncpy(genOpts.dynamic_uri, optarg, MAX_URI);
      break;
      
    case 20:
      genOpts.static_only = 1;
      break;

    case 21:
      genOpts.num_clients = atoi(optarg);
      break;
    
    default:
      printf("Unrecognized option %c\n", c);
      usage(argv[0]);
    }
  }
}




/********************************************************************
 * Initialize the tracking arrays for each file type and file class.
 * These arrays are used to track how many files of each 
 * type and class are generated
 * Requires the num_dirs global to have been set! 
 ********************************************************************/
void init() {
  int i;
  double tableSum = 0.0;
  double cumulative = 0.0;

  for( i = 0; i <= MAX_FILE; i++ ) {
    fileDist[i] = 0.0;
  }
  
  for( i = 0; i <= MAX_CLASS; i++ ) {
    classDist[i] = 0.0;
  }

  for( i = 0; i < NUM_REQ_TYPES; i++ ) {
    requestDist[i] = 0.0;
  }

#ifdef USE_LONG_SESSIONS
  sessionPDF[MAX_SESSION] = 1.0;
  for( i = 0; i < MAX_SESSION; i++ ) {
    sessionPDF[i] = 0.0;
  }
#endif /* USE_LONG_SESSIONS */

  for( i = 0; i <= MAX_SESSION; i++ ) {
    sessionDist[i] = 0.0;
  }

  if( genOpts.num_dirs > 0 ) {
    /* Allocate an array to track the directory distribution */
    directoryDist = (double*) malloc( sizeof(double) * genOpts.num_dirs );
    /* Generate the discrete zipf distribution for the 
     * directory access pattern 
     */
    dirCDF = (double*) malloc( sizeof(double) * genOpts.num_dirs );
    dirPDF = (double*) malloc( sizeof(double) * genOpts.num_dirs );
    dirPDF[0] = 0.0;

    for( i = 0; i < genOpts.num_dirs; i++ ) {
      directoryDist[i] = 0.0;
      dirPDF[i] = ((double) 1.0) / ((double) i+1);
      tableSum += dirPDF[i];
    }

    for( i = 0; i < genOpts.num_dirs; i++ ) {
      dirPDF[i] = (dirPDF[i] / tableSum);
      cumulative += dirPDF[i];
      dirCDF[i] = cumulative;
    }
    /* for */
  } /* if */  
  
}


/*************************************************************************
 * Determine if the distribution of requests represented by dist matches
 * the probability density function specified in pdf within a factor of
 * epsilon. The request distribution in dist must be normalized using
 * the total number of requests (totalReqs). Return 1 if
 * the distribution matches within the specified tolerance, 0 otherwise.
 *************************************************************************/
int tolerance_met(double * dist, double * pdf, int size, 
                   double epsilon, double totalReqs ) {
  int index;
  double fraction = 0.0;
  double value = 0.0;

  for( index = 0; index < size; index++ ) {

    if( pdf[index] == 0.0 ) {
     if(dist[index] != 0.0) return 0;

    } else {

      fraction = dist[index] / totalReqs;
      value = fabs(pdf[index] - fraction) / pdf[index];
      if( value > epsilon ) return 0;
    }
  } /* for */ 

  return 1;
}


/***************************************************************************
 * Write the given distribution out to a file.
 ***************************************************************************/
void writeDist(FILE * sf, double *dist, double *pdf, int size, double epsilon, double totalReqs)
{
  double percent = 0.0;
  int i;
  
  fprintf(sf, "# %5s %8s %10s %10s %10s %5s\n",
          "Index", "Count", "Actual%", "Target%", "Diff%", "Met");
  for (i=0; i<size; i++) {
    percent = dist[i] * 100.0 / (double) totalReqs;
    if ((dist != sessionDist) || (percent != 0.0)) {
      fprintf(sf, "# %5d %8.0lf %10.6lf %10.6lf %10.6lf %5s\n", 
	      i, dist[i], percent, pdf[i] * 100.0, 
	      fabs(percent - (pdf[i] * 100.0)), 
	      tolerance_met(dist, pdf, i, epsilon, totalReqs) ? "Yes" : "No");
    }
  }
}


/*************************************************************
 * Remove unused sessions from the distribution. This allows
 * us to determine if the sessions that were written to the log
 * file(s) meet the desired tolerance.
 *************************************************************/
void distribution_remove_unused() {
  int i,j,k;
  session * s = NULL;
  uri * u;
  int length;
  
  /* Process sessions of every length */
  for( i = 0; i < MAX_SESSION; i++ ) {
    /* Process all unused sessions of this length */
    length = sessions[i].length;
    for( j = 0; j < length; j++ ) {
      s = remove_first_session(&(sessions[i]));
      assert(s != NULL);
      /* Process every URI within an unused session */
      for( k = 0; k < s->length; k++ ) {
	u = &(s->uris[k]);
	fileDist[u->file]--;
	classDist[u->class]--;
	directoryDist[u->dir]--;
      } /* for k */
      sessionDist[s->length]--;
      free_session(s);
    } /* for j */    
  } /* for i */
}


/*****************************************************************
 * Record the specified triple (file,class,dir) so that we
 * can periodically check the distribution of generated objects 
 * (files, classes, and directories) against the required 
 * (SPECweb99) distributions
 *****************************************************************/
void distribution_add(int file, int class, int dir, req_type req) {

  assert( file  <= MAX_FILE );
  assert( class <= MAX_CLASS );
  assert( dir   <  genOpts.num_dirs );

  fileDist[file]++;
  classDist[class]++;
  directoryDist[dir]++;
  requestDist[req]++;
}


/******************************************************************* 
 * We've just created a request for a this file and class.
 * Track the current distribution and when we have reached the
 * target distribtion for all files and all classes we've got
 * a set of sessions that meets the requirements.
 *
 * When we've met the requirements return 1.
 * Otherwise return 0.
 * We Want to ensure that:
 *   sessions of length = 1    = 30%
 *   sessions of length = 5-15 = 70%
 *   sessions of length 5 - 15 are roughly uniform (average is 10)
 *
 *   class 0  = 35 %
 *   class 1  = 50 %
 *   class 2  = 14 %
 *   class 3  =  1 %
 *
 *   file  0  =  3.9 %
 *   file  1  =  5.9 %
 *   file  2  =  8.8 %
 *   file  3  = 17.7 %
 *   file  4  = 35.3 %
 *   file  5  = 11.8 %
 *   file  6  =  7.1 %
 *   file  7  =  5.0 %
 *   file  8  =  4.4 %
 *******************************************************************/ 
int distribution_check( double epsilon ) {
  int i;

  /* Check the file distribution */
  if (!tolerance_met(fileDist, filePDF, MAX_FILE + 1, epsilon, totalRequests)) {
    ldbg("distribution_check: doesn't meet file "
         "distribution tolerances\n");
    return 0;
  }
  
  /* Now check the class distribution */
  if (!tolerance_met(classDist, classPDF, MAX_CLASS + 1, epsilon, totalRequests))  {
    ldbg("distribution_check: doesn't meet "
	 "class distribution tolerances\n");
    return 0;
  }
  
  /* Now check the directory number distribution*/
  if (!tolerance_met(directoryDist, dirPDF, genOpts.num_dirs, epsilon, totalRequests)) {
    ldbg("distribution_check: doesn't meet directory "
	 "distribution tolerances\n");
    return 0;
  }

  /* We won't meet request-type distributions if we're only generating one kind of request */
  if( genOpts.static_only == 0 && genOpts.dynamic_only == 0 ) {
    /* Check the distribution of request types */
    if (!tolerance_met(requestDist, reqTypePDF, NUM_REQ_TYPES, epsilon, totalRequests)) {
      ldbg("distribution_check: doesn't meet request "
	 "distribution tolerances\n");
      return 0;
    }
  }
 
  ldbg("distribution_check: all distribution tolerances met\n");
  return 1;
}



/******************************************************************
 * Generate a value according to the given CDF vector.
 * This function uses the uniform distribution to generate integers
 * that adhere to the given CDF 
 ******************************************************************/
void genValue( double * CDF, int size, int * value ) {
  int i;
  double random;

  random = ((double) rand()) / ((double) RAND_MAX);
  for( i = 0; i < size; i++ ) {
    if( random <= CDF[i] ) {
      *value = i;
      return;
    }
  }
  *value = (size - 1);
}


/************************************************************** 
 * Write the given URI to the specified file. Assume that the
 * file has already been opened, and is ready for writing
 **************************************************************/
void write_uri( uri * u, FILE * log ) {

  assert( u != NULL );
  assert( log != NULL );
  if( u->type == STATIC ) {
    fprintf(log, "%s%s/dir%05d/class%1d_%1d think=%1.3f\n", genOpts.root_dir, SPECPATH, u->dir, u->class, u->file, u->think);

  } else if ( u->type == DYNAMIC_GET ) {
    fprintf(log, "%s%s%s/dir%05d/class%1d_%1d think=%1.3f\n", genOpts.dynamic_uri, genOpts.root_dir, 
	    SPECPATH, u->dir, u->class, u->file, u->think);

  } else if ( u->type == DYNAMIC_GET_AD ) {
    assert( u->mycookie != NULL );
    fprintf(log, "%s%s%s/dir%05d/class%1d_%1d cookie=\"my_cookie=user_id=%d&last_ad=%d\" think=%1.3f\n", genOpts.dynamic_uri, genOpts.root_dir, 
	    SPECPATH, u->dir, u->class, u->file, u->mycookie->userid, u->mycookie->lastad, u->think);

  } else if ( u->type == DYNAMIC_POST ) {
    fprintf(log, "%s method=POST contents=\"urlroot=%s%s/&dir=%05d&class=%1d&num=%1d&client=%d\" think=%1.3f\n", genOpts.dynamic_uri, genOpts.root_dir,
	    SPECPATH, u->dir, u->class, u->file, u->mycookie->userid, u->think);

  } else {
    assert(0); /* Unknown URI type */
  }
}


/************************************************************** 
 * Return a req_type based on the SPECweb99 distributions
 * According to SPEC, 70% of requests are static, 12.45%
 * are  Standard Dynamic GET, etc. See the SPECweb99
 * Design Document for details.
 **************************************************************/
req_type gen_req_type() {
  double random;
  int i;

  if( genOpts.static_only ) return STATIC;

  random = ((double) rand()) / ((double) RAND_MAX);

  /* Only generate dynamic requests */
  if( genOpts.dynamic_only ) {

    for( i = 0; i < NUM_REQ_TYPES; i++ ) {
      if( random <= dynamicReqTypeCDF[i] ) {
        break;
       }    
    } /* for */ 
    
  } else {

    for( i = 0; i < NUM_REQ_TYPES; i++ ) {
      if( random <= reqTypeCDF[i] ) {
        break;
       }    
    } /* for */ 

  }

  return (req_type) i;
}

/**************************************************************
 * Generate a cookie for a SPECweb99 Dynamic GET with ad 
 * rotation. The cookie contains a userid and a last_ad
 * field. Those are generated here.
 * Return NULL if an error occurs, otherwise return
 * a pointer to the cookie
 **************************************************************/

cookie * generate_cookie() {
  cookie * c = NULL;

  c = (cookie*) malloc( sizeof(cookie) );
  if ( c == NULL ) return NULL;
  /* This is a hack, and doesn't really model SPECweb99
     For the userid and last_ad we use uniform distributions */
  /* Also the SPECweb99 appserver expects userid > 10000 */
  c->userid = 10000 + (rand() % (genOpts.num_clients + 1));
  c->lastad = rand() % (MAX_AD_NUMBER + 1);
  return c;
}



/******************************************************************
 * Fill in the given session data structure with URIs
 * <sess> is a pointer to the empty session data structure
 * <length> is the number of URIs the session can hold 
 ******************************************************************/
void fill_session( session * sess, int length ) {
  uri * u;
  int i;
  int web_objects = 0; /* How many embedded objects in a page */

  for( i = 0; i < length; i++ ) {    

    assert(sess != NULL);
    u = &(sess->uris[i]);

    /* What kind of URI? Static? Dynamic? etc. */
    u->type = gen_req_type();

    if( u->type == DYNAMIC_GET_AD || u->type == DYNAMIC_POST ) {
      /* Generate a cookie for Dynamic GET with ad rotation or POST requests 
         For POST requests, only the userid part of the cookie is actually used */
      u->mycookie = generate_cookie();
    }

    /* Generate a (file,class,dir) triple to represent the URI */
    genValue(fileCDF, (MAX_FILE+1), &(u->file));
    genValue(classCDF, (MAX_CLASS+1), &(u->class));
    genValue(dirCDF, genOpts.num_dirs,&(u->dir) );    
    distribution_add(u->file, u->class, u->dir, u->type);

    /* Each page has a number of embedded web objects. How many? */
    if( web_objects == 0 ) {
      genValue(webObjectsCDF, (MAX_SESSION+1),  &web_objects);
      u->think = PARETO_MEAN;
    } else {
      u->think = WEIBULL_MEAN;
      web_objects--;
    }
    /* Overwrite the Weibull/Pareto think times if we are using fixed thinks */
    if( genOpts.fixed_think >= 0.0 ) {
      u->think = genOpts.fixed_think;
    }
    
    total_think_time += u->think;
    sess->length++;
    sess->bytes += spec_file_sizes[u->class*9 + u->file];
    totalRequests++;
  }
}

/***********************************************************************
 * Generate uris that prime the cache 
 ***********************************************************************/
void add_uri_prime( session * sess ) {
  uri * u;
  static int last_file = 0;
  static int last_class = 0;
  static int last_dir = 0;
  
  assert(sess != NULL);
  if (last_file > MAX_FILE) {
    last_file = 0;
    last_class++;
  }
  if (last_class > MAX_CLASS) {
    last_class = 0;
    last_dir++;
  }
  if (last_dir > genOpts.num_dirs) {
    last_dir = 0;
  }

  // There's really no hope of matching the distributions this way
  //distribution_add(last_file, last_class, last_dir);
  u = &(sess->uris[sess->length]);
  u->file = last_file;
  u->class = last_class;
  u->dir = last_dir;
  u->think = 0.0;
  sess->length++;
  last_file++;
}


/*********************************************************************
 * Generate a session length according to the SPECweb99 session length
 * distributions. HTTP 1.0 sessions are simulated by using HTTP 1.1
 * sessions, but only sending a single request over the connection.
 *********************************************************************/
int get_session_len() {
#ifdef USE_LONG_SESSIONS
  return (LONG_SESSION_LENGTH);
#else
  int session_len = 0;
  double random;
  random = ((double) rand()) / ((double) RAND_MAX);
  
  // 30% of connections are HTTP 1.0. 
  // The remaining 70% are HTTP 1.1
  if( random <= 0.3 ) {
    // Simulate an HTTP 1.0 connection by using an
    // HTTP 1.1 session with just 1 request per connection
    session_len = 1;
  } else {
    // HTTP 1.1 session with multiple requests per connection
    random = (((double) rand()) / ((double) RAND_MAX)) * 11.0;
    session_len = floor(5.0 + random);
  }
  return session_len;
#endif /* USE_LONG_SESSIONS */
}


/*********************************************************************
 * Add a session of the given length to the session tracking array
 *********************************************************************/
void session_add(int session_len ) {
  ldbg("session_add: len = %d\n", session_len);
  if( !genOpts.prime_cache) assert(session_len == 1 || (session_len >= 5 && session_len <= 15));  
  sessionDist[session_len]++;  
}



/*********************************************************************
 * Check the distribution of session-lengths for the generated
 * sessions against the SPECweb99 specified session-length distribution.
 * That is, make sure that we have the right number of sessions of each
 * length.
 *********************************************************************/
int session_check(double epsilon ) {
  int i;
  
  /* For each session length,  */
  for( i = 0; i <= MAX_SESSION; i++ ) {      
    if( sessionPDF[i] > 0.0 ) {
      if (!tolerance_met(sessionDist, sessionPDF, i, epsilon, session_count)) return 0;
    }
  }
    
  ldbg("session_check: session distribution met\n");
  return 1;
}


void writeStatsFile( FILE * sf, int argc, char * argv[], log_stats * stats ) {
  int i;
  unsigned int total_reqs = 0;
  unsigned int total_sess = 0;
  double total_bytes = 0.0;


  fprintf(sf, "# Command Line:\n#    ");
  for( i = 0; i < argc; i++ ) {
    fprintf(sf, "%s ", argv[i] );
  }
  fprintf(sf, "\n#\n");
  fprintf(sf, "# Random Seed = %d\n#\n", genOpts.random_seed );  
  fprintf(sf, "#  %4s   %8s  %8s  %10s  %16s  %12s  %12s  %11s\n", "Log#", "Sessions", "Requests",
	  "Reqs/sess", "Bytes Requested", "Bytes/sess", "Bytes/req", "Total think");
  for( i = 0; i < genOpts.num_logs; i++ ) {
    if( stats[i].sessions > 0 && stats[i].requests > 0 ) {
      fprintf(sf, "# %5d   %8d  %8d  %10.2f  %14.0f   %12.2f   %12.2f  %12.2f\n", i+1, stats[i].sessions,
	      stats[i].requests, 
	      (double)stats[i].requests / (double) stats[i].sessions,
	      stats[i].bytes, 
	      (double)(stats[i].bytes / (double)(stats[i].sessions)),
	      (double)(stats[i].bytes / (double)(stats[i].requests)),
	      stats[i].think_time);
      total_reqs += stats[i].requests;
      total_sess += stats[i].sessions;
      total_bytes += stats[i].bytes;
    }    
  } /* for */

  fprintf(sf, "# %7s  %8d %8d  %10.2f  %14.0f   %12.2f   %12.2f\n\n", "TOT/AVG", total_sess,
	  total_reqs, 
	  (double)total_reqs / (double) total_sess,
	  total_bytes,
	  (double)(total_bytes / ((double)total_sess)),
	  (double)(total_bytes / ((double)total_reqs)));
  
  fprintf(sf, "# Total sessions generated = %d\n", session_count);
  fprintf(sf, "# Total requests generated = %ld\n", (long int) totalRequests);
  fprintf(sf, "# Sessions written to file = %d\n", total_sess);
  fprintf(sf, "# Requests written to file = %d\n", total_reqs);
  fprintf(sf, "# Discarded sessions = %d\n", (session_count - total_sess));
  fprintf(sf, "# Discarded requests = %d\n#\n", (int) totalRequests - total_reqs);
  fprintf(sf, "# AGGREGATE DISTRIBUTIONS (excluding discarded sessions)\n");
  fprintf(sf, "# -----------------------------------------------------\n");
  fprintf(sf, "# File Distributions\n");
  writeDist(sf, fileDist, filePDF, MAX_FILE+1, genOpts.tolerance, totalRequests );
  fprintf(sf,"#\n# Class Distributions\n");
  writeDist(sf, classDist, classPDF, MAX_CLASS+1, genOpts.tolerance, totalRequests);
  fprintf(sf, "#\n# Directory Distributions\n");
  writeDist(sf, directoryDist, dirPDF, genOpts.num_dirs, genOpts.tolerance, totalRequests);
  fprintf(sf, "#\n# Session Distributions\n");
  writeDist(sf, sessionDist, sessionPDF, MAX_SESSION+1, genOpts.tolerance, session_count );
  fprintf(sf, "#\n#\n");
  fprintf(sf, "#\n# Request Type Distributions\n");
  writeDist(sf, requestDist, reqTypePDF, NUM_REQ_TYPES, genOpts.tolerance, totalRequests );
  fprintf(sf, "#\n#\n");
  fflush(sf);
}


/* Split the generated sessions across all available log files */
void write_log_files(int argc, char * argv[] ) {
  int i,j;
  FILE * log; /* Log file */
  FILE * sf;  /* Temporary stats file */
  char cmd[128];
  char filename[MAX_FILE_CHARS];
  int sessions_per_file[MAX_SESSION];
  int sessions_added[MAX_SESSION];
  int done = 0;
  int sess_len;
  session * sess;
  uri * u;
  log_stats stats[genOpts.num_logs];

  /* Calculate how many sessions of each length go into each file */
  for( i = 0; i < MAX_SESSION; i++ ) {
    sessions_per_file[i] = sessions[i].length / genOpts.num_logs;
  }

  for( i = 0; i < genOpts.num_logs; i++ ) {
    /* First open the log file */
    sprintf(filename, "%s-%d.log", LOGFILE_ROOT, i+1);
    log = fopen(filename, "w" );
    if( log == NULL ) {      
      perror("Could not open client log file for writing\n");
      return;
    }

    /* Track how many session of each length have been added to this file */
    for(j = 0; j < MAX_SESSION; j++ ) {
      sessions_added[j] = 0;
    }
  
    /* Zero the stats for this log */  
    stats[i].sessions = 0;
    stats[i].requests = 0;
    stats[i].bytes = 0.0;
    stats[i].think_time = 0.0;

    /* Write all the sessions for this log file */
    done = 0;
    while( !done ) {

      /* What length of session do we add? */
      sess_len = get_session_len() - 1;
      if(sessions_added[sess_len] <  sessions_per_file[sess_len] ) {
	/* Remove this session from the list of available sessions */
	sess = remove_first_session( &(sessions[sess_len]) );
	if( sess == NULL ) continue;
	fprintf(log, "\n#Session #%d (length = %d, bytes = %d)\n", 
		++(stats[i].sessions), sess->length, sess->bytes);

	/* Write out each request in the session */
	for(j=0; j < sess->length; j++  ) {
	  u = &(sess->uris[j]);
	  stats[i].think_time += u->think;
	  write_uri(u, log);    
	}
	/* Update the stats for this log */
	stats[i].requests += sess->length;
	stats[i].bytes += (double) sess->bytes;
	sessions_added[sess_len]++;
	free_session(sess);
      }

      /* Decide if we're done or not */
      done = 1;
      for( j = 0; j < MAX_SESSION; j++ ) {
	if( sessions_added[j] < sessions_per_file[j] )
	  done = 0;
      }
    } /* while */
    
    /* Make sure to close the log file */
    fclose(log);   
  } /* for */  

  printf("Discarding unused sessions...\n");
  distribution_remove_unused();
  /* Write statistics to a separate file */
  sf = fopen(STATSFILE, "w");
  if (sf == NULL) {
    perror("Unable to open stats file. No statistics produced\n");
    exit(1);
  }  
  writeStatsFile(sf, argc, argv, stats);  
  for( i = 0; i < genOpts.num_logs; i++ ) {
    /* Prepend the stats file to the log file */
    sprintf(filename, "%s-%d.log", LOGFILE_ROOT, i+1);
    snprintf(cmd, 128, "/bin/cat %s %s > %s; /bin/mv %s %s", 
	     STATSFILE, filename, TEMPFILE, TEMPFILE, filename);
    system(cmd);
  }
  snprintf(cmd, 128, "/bin/rm -f %s", STATSFILE );
  system(cmd); 
}


/* Tar and zip the log files for easy transportation */
int zip_log_files(char * zipfile ) {
  char * cmd = NULL;
  int i, len, offset;

  len = genOpts.num_logs * strlen(LOGFILE_ROOT) + 128;
  cmd = (char*) malloc( len );
  if( cmd == NULL ) return 0;
  cmd[len-1] = '\0';
  offset = snprintf(cmd, len-1, "%s %s ", "tar zcvf", zipfile);

  for( i = 0; i < genOpts.num_logs; i++ ) {
    offset += sprintf(cmd+offset, "%s-%d.log ", LOGFILE_ROOT, i+1);
  }
  system(cmd);
  free(cmd);
}



/****************************************************************************************/
/* 					MAIN 						*/
/****************************************************************************************/
int main(int argc, char *argv[])
{
  int i;
  int done = 0;
  int session_len = 0;
  int dist_met = 0;
  int session_dist_met = 0;
  session * sess;

  initOpts();
  process_args(argc, argv);
  checkOpts();
  
  // Seed the random number generator and initialize the tuple-tracking arrays
  if (genOpts.random_seed == 0) {
    genOpts.random_seed = (unsigned int) time(NULL);
  } 
  srand(genOpts.random_seed);
  init();
  
  printf("Generating sessions...\n");
  while (session_count <= genOpts.max_sessions && !done) {    

    session_count++;        
    session_len = get_session_len();   
    sess = new_session(session_len);    
    fill_session( sess, session_len );	      
    assert(sess->length > 0);
    add_session(&(sessions[sess->length - 1]), sess );
    
    dist_met = distribution_check(genOpts.tolerance - genOpts.delta);
    session_add( session_len );
    /* Need to check the session distribution */
    session_dist_met = session_check(genOpts.tolerance - genOpts.delta);
    if (session_dist_met && dist_met && (session_count > genOpts.min_sessions)) {
      done = 1;
    }
  } /* while */  
  
  printf("Generated %d sessions\n", session_count );
  if (! (dist_met && session_dist_met)) {
    printf("WARNING: distribution tolerances have not been met\n\n");
  }  
  if (dist_met && session_dist_met && (session_count > genOpts.min_sessions)) {
    printf("Distribution tolerances and minimum sessions were met\n");
  }  
  if (session_count > genOpts.max_sessions) {
    printf("Program terminated: max_sessions was exceeded\n\n");     
  }
  
  printf("Writing sessions to %d different log files...\n", genOpts.num_logs);
  write_log_files(argc, argv); 
  
  printf("Re-checking distribution tolerances and minimum sessions\n");
  dist_met = distribution_check(genOpts.tolerance - genOpts.delta);
  session_dist_met = session_check(genOpts.tolerance - genOpts.delta);
  if (! (dist_met && session_dist_met)) {
    printf("WARNING: Too many sessions were discarded. Tolerances no longer satisfied\n");
  } else {
    printf("Tolerances are still satisfied in spite of discarded sessions\n");
  } 
  if( genOpts.zip ) {
    printf("Creating archive %s\n", genOpts.zip_file);
    zip_log_files(genOpts.zip_file);
  }

  free( dirCDF );
  free( dirPDF );
  free( directoryDist );
  //  print_stats();
  exit(0);
}
