/*
    userver -- (pronounced you-server or micro-server).
    This file is part of the userver, a high-performance web server designed for
    performance experiments.
          
    This file is Copyright (C) 2005  David Pariag

    Authors: David Pariag <db2paria@cs.uwaterloo.ca>
    See AUTHORS file for list of contributors to the project.
  
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/


/*
 * Title   : http_log_gen.c
 * Author  : David Pariag
 * Date    : Feb 4th 2005
 *
 */

/*
 * Size information for SpecWeb99 file set
 *             Class   ClassSize  Cumulative
 *     102      0.00                     102
 *     204      0.00                     306
 *     307      0.00                     613
 *     409      0.00                    1022
 *     512      0.00                    1534
 *     614      0.00                    2148
 *     716      0.00                    2864
 *     819      0.00                    3683
 *     921      0.00        4604        4604
 *    1024      1.00                    5628
 *    2048      1.00                    7676
 *    3072      1.00                   10748
 *    4096      1.00                   14844
 *    5120      1.00                   19964
 *    6144      1.00                   26108
 *    7168      1.00                   33276
 *    8192      1.00                   41468
 *    9216      1.00       46080       50684
 *   10240      2.00                   60924
 *   20480      2.00                   81404
 *   30720      2.00                  112124
 *   40960      2.00                  153084
 *   51200      2.00                  204284
 *   61440      2.00                  265724
 *   71680      2.00                  337404
 *   81920      2.00                  419324
 *   92160      2.00      460800      511484
 *  102400      3.00                  613884
 *  204800      3.00                  818684
 *  307200      3.00                 1125884
 *  409600      3.00                 1535484
 *  512000      3.00                 2047484
 *  614400      3.00                 2661884
 *  716800      3.00                 3378684
 *  819200      3.00                 4197884
 *  921600      3.00     4608000     5119484
 *
 * 5119484               5119484
 */

#ifndef HTTP_LOG_GEN
#define HTTP_LOG_GEN


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <math.h>
#include <time.h>
#include <limits.h>
#include <assert.h>
#include "session.h"

/* #define LOCAL_DEBUG */

#ifdef LOCAL_DEBUG
#define ldbg(...) \
  do {\
    printf(__VA_ARGS__); \
    fflush(stdout); \
    fflush(stderr); \
  } while (0);
#else
#define ldbg(x, ...)
#endif

#define LOGFILE_ROOT             "httpspec99"
#define PRIME_LOGFILE_NAME       "httpspec99-prime.log"
#define PRIME_CLEAR_LOGFILE_NAME "httpspec99-prime-clear.log"
#define STATSFILE                "/tmp/statsfile"
#define TEMPFILE                 "/tmp/spec.log"
#define REQUESTDIR               "/specweb99"
#define SPECPATH                 "/file_set"
#define DYNAMIC_SCRIPT           "/specweb99-fcgi.pl?"

//#define USE_LONG_SESSIONS
#define LONG_SESSION_LENGTH        (741)

#define MAX_URI        (512)
#define MAX_DIR_LEN    (512)
#define MAX_FILE_CHARS (200)
#define MAX_CLASS      (3)
#define MAX_FILE       (8)
#define MIN_SESSION    (1)
#define NUM_SPEC_FILES  36

#ifdef USE_LONG_SESSIONS
#define MAX_SESSION    (LONG_SESSION_LENGTH)
#else
#define MAX_SESSION    (15)
#endif /* USE_LONG_SESSIONS */

#define FILES_PER_DIR  ((MAX_FILE+1) * (MAX_CLASS+1))
#define BYTES_PER_DIR  (5119484)
#define MBS_PER_DIR    (4.88232040405273437500)

#define MAX_THINK (1.0)

/* Max allowable length for a URI.
   The longest known URIs (Dynamic GET) may be 67 chars long*/
#define URI_LEN 128

#define MAX_AD_NUMBER 359

/**************************************************** 
 * CDF (discrete distribution) for various kinds of 
 * files in the SpecWeb fileset. 
 * See SPECweb99 "Design Document" for details
 ****************************************************/

/* Describe the breakdown among different request types 
   (static, dynamic get, dynamic get with ad rotation, dynamic post) */
static double reqTypePDF[NUM_REQ_TYPES] = {
  0.70, 0.126, 0.126, 0.048
};

static double reqTypeCDF[NUM_REQ_TYPES] = {
  0.70, 0.826, 0.952, 1.0
};

/* Used to generate dynamic-only logs. Scale all the
   dynamic percentages to compensate for the lack of static
 */
static double dynamicReqTypeCDF[NUM_REQ_TYPES] = {
  0.00, 0.415, 0.835, 1.0
};

static double fileCDF[MAX_FILE+1] = {
  0.039, 0.098, 0.186, 0.363, 0.716, 0.834, 0.905, 0.955, 1.0
};


/* PDF (discrete distribution) for various kinds of 
 * files in the SpecWeb fileset 
 */
static double filePDF[MAX_FILE+1] = {
  0.039, 0.059, 0.088, 0.177, 0.353, 0.118, 0.071, 0.05, 0.044
};


/* CDF (discrete distribution) for file classes in the SpecWeb fileset.*/
static double classCDF[MAX_CLASS+1] = {
  0.35, 0.85, 0.99, 1.0
};

/* CDF (discrete distribution) for the file classes in the SpecWeb fileset*/
static double classPDF[MAX_CLASS+1] = {
  0.35, 0.50, 0.14, 0.01
};

/*********************************************************************
 * CDF describing the distribution of the number of embedded objects 
 * (e.g. images) in a web page. This distribution determines if think 
 * times are distributed using a Pareto or Weibull distribution. 
 * This CDF was obtained from Barford and Crovella's SURGE paper
 *********************************************************************/
static double webObjectsCDF[MAX_SESSION+1] = {
  0.0, 0.0,  0.8744, 0.9307, 0.9656, 0.9799, 0.9871, 0.9912, 0.9936,
  0.9941, 0.9954, 0.9963, 0.997, 0.9975, 0.9979, 1.0
};


#ifdef USE_LONG_SESSIONS
static double sessionPDF[MAX_SESSION+1];
#else
static double sessionPDF[MAX_SESSION+1] = {
  0.0, 0.3, 0.0, 0.0, 0.0, 0.0636, 0.0636, 0.0636, 0.0636, 
  0.0636, 0.0636, 0.0636, 0.0636, 0.0636, 0.0636, 0.0636
};

#endif /* USE_LONG_SESSIONS */


/****************************************************** 
 * Hardcode the size of the SPECweb99 files.
 * This allows us to calculate the number of bytes
 * in a given session
 ******************************************************/
static int spec_file_sizes[NUM_SPEC_FILES] = {
  102, 204,307,409,512,614,716,819,921,1024,2048,3072,4096,
  5120,6144,7168,8192,9216,10240,20480,30720,40960,51200,
  61440,71680,81920,92160,102400,204800,307200,409600,
  512000,614400,716800,819200,921600
};


/************************************************************** 
 * These will eventually point to arrays that will hold the 
 * CDF and PDF respectively. The size of the array depends on 
 * the max_dir parameter, so we can't allocate arrays a priori 
 **************************************************************/
static double * dirCDF;
static double * dirPDF;

/* Track how many files of each type are actually generated in our workload.*/
static double fileDist[MAX_FILE+1];

/* Track how many files of each class are actually generated in our workload.*/
static double classDist[MAX_CLASS+1];

/* Track how many sessions of each length are 
 * actually generated in the workload 
 */
static double sessionDist[MAX_SESSION+1];

/* Track how many times each directory appears in the workload */
static double * directoryDist;

/* Track request types that are generated */
static double requestDist[NUM_REQ_TYPES];

/* Track the total number of requests generated */
static double totalRequests = 0.0;

/* Total number of sessions in the workload*/
static int session_count = 0;

/* Stats relating to user think times */
static double total_think_time = 0.0;


typedef struct opts {
  int clear_stats;
  int num_dirs;
  int num_clients;
  int timer_esc;
  int prime_cache;
  int min_sessions;
  int max_sessions;
  int fileset_size;
  float tolerance;
  float delta;  
  int use_think_times;
  int dynamic_only;
  int static_only;
  int no_http_10;
  unsigned int random_seed;
  int num_logs;
  int zip;
  char zip_file[MAX_URI];
  float fixed_think;
  char root_dir[MAX_DIR_LEN];
  char dynamic_uri[MAX_URI];
} opts;

/* Use this struct to gather stats about each log file. */
typedef struct log_stats {
  int sessions;      /* Number of sessions in the log file */
  int requests;      /* Number of requests in the log file */
  double bytes;      /* Number of bytes requested in the log file */
  double think_time; /* Sum of think times across all sessions in a log */
} log_stats;
#endif /* HTTP_LOG_GEN */
