/*
    userver -- (pronounced you-server or micro-server).
    This file is part of the userver, a high-performance web server designed for
    performance experiments.
          
    This file is Copyright (C) 2005  David Pariag

    Authors: David Pariag <db2paria@cs.uwaterloo.ca>
    See AUTHORS file for list of contributors to the project.
  
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/


/* Title:  session.c
 * Author: David Pariag
 * Date:   Feb 4th 2005
 */


#include "session.h"
#include "http_log_gen.h"

static int created = 0;
static int freed = 0;
static int added = 0;
static int removed = 0;

/********************************************************
 * Create and initialize a new session structure.
 * The length parameter indicates the number of requests
 * in the session. 
 *
 * Returns: A pointer to the allocated session, or NULL
 * if an error occured
 ********************************************************/
session * new_session(int length) {
  session * sess = NULL;
  int i = 0;  

  sess = (session*) malloc(sizeof(session));
  if( sess == NULL ) return NULL;
  // Only set the length once the URIs are actually in place
  sess->length = 0;
  sess->bytes = 0;
  sess->next = NULL;
  sess->uris = (uri*) malloc( sizeof(uri) * length );
  if( sess->uris == NULL ) {
    free(sess);
    return NULL;
  }
  /* Initialize the uri array with sentinel values */
  for( i = 0; i < length; i++ ) {
    sess->uris[i].dir = -1;
    sess->uris[i].class = -1;
    sess->uris[i].file = -1;
    sess->uris[i].think = -1.0;
    sess->uris[i].mycookie = NULL;
    sess->uris[i].type = NUM_REQ_TYPES;
  }
  created++;
  return sess;
}


/********************************************************
 * Free (deallocate) the memory associated with the given
 * session.
 ********************************************************/
void free_session(session * tofree ) {
  int i = 0;

  assert(tofree != NULL);
  
  for( i = 0; i < tofree->length; i++ ) {
    if( tofree->uris[i].mycookie != NULL ) {
      free( tofree->uris[i].mycookie);
    }
  } /* for */

  free( tofree->uris );
  tofree->next = NULL;
  free( tofree );  
  freed++;
}


/********************************************************
 * Add the given session to front of the specified 
 * session list
 ********************************************************/
void add_session(session_list * s,  session * toadd ) {  
  assert(toadd != NULL);
  assert(s != NULL);
  
  if( s->head == NULL ) {
    s->head = toadd;
    toadd->next = NULL;
    
  } else {    
    toadd->next = s->head;
    s->head = toadd;
  }
  s->length++;
  added++;
}

/********************************************************
 * Remove and return the first session from the 
 * given session_list
 *
 * Returns: A pointer to the removed session, or NULL
 * if the given list is empty
 ********************************************************/
session * remove_first_session(session_list * s) {
  session * first;
  
  assert(s != NULL);
  if( s->head == NULL || s->length == 0 ) return NULL;
  
  first = s->head;
  s->head = s->head->next;
  s->length--;
  removed++;
  return first;
}


void print_stats() {
  printf("Sessions created: %d\n", created);
  printf("Sessions freed: %d\n", freed);
  printf("Sessions added: %d\n", added);
  printf("Sessions removed: %d\n", removed);
}
